from dataclasses import dataclass, field
from typing import List

from emptylog.call_data import LoggerCallData


@dataclass
class LoggerAccumulatedData:
    debug: List[LoggerCallData] = field(default_factory=list)
    info: List[LoggerCallData] = field(default_factory=list)
    warning: List[LoggerCallData] = field(default_factory=list)
    error: List[LoggerCallData] = field(default_factory=list)
    exception: List[LoggerCallData] = field(default_factory=list)
    critical: List[LoggerCallData] = field(default_factory=list)

    def __len__(self) -> int:
        return len(self.debug) + len(self.info) + len(self.warning) + len(self.error) + len(self.exception) + len(self.critical)



from abc import ABC

from emptylog.protocols import LoggerProtocol


class AbstractLogger(LoggerProtocol, ABC):
    def __repr__(self) -> str:
        return f'{type(self).__name__}()'

    def __add__(self, other: LoggerProtocol) -> 'LoggersGroup':  # type: ignore[name-defined] # noqa: F821
        if not isinstance(other, LoggerProtocol):
            raise NotImplementedError('The addition operation is defined only for loggers.')

        from emptylog import LoggersGroup

        local_loggers = self.loggers if isinstance(self, LoggersGroup) else [self]
        other_loggers = other.loggers if isinstance(other, LoggersGroup) else [other]

        return LoggersGroup(*local_loggers, *other_loggers)

    def __radd__(self, other: LoggerProtocol) -> 'LoggersGroup':  # type: ignore[name-defined] # noqa: F821
        if not isinstance(other, LoggerProtocol):
            raise NotImplementedError('The addition operation is defined only for loggers.')

        from emptylog import LoggersGroup

        local_loggers = self.loggers if isinstance(self, LoggersGroup) else [self]
        other_loggers = other.loggers if isinstance(other, LoggersGroup) else [other]

        return LoggersGroup(*other_loggers, *local_loggers)



from emptylog import MemoryLogger, LoggerProtocol
from emptylog.memory_logger import LoggerCallData


def test_memory_logger_is_logger():
    assert isinstance(MemoryLogger(), LoggerProtocol)


def test_memory_logger_is_working():
    attribute_names = [
        'debug',
        'info',
        'warning',
        'error',
        'exception',
        'critical',
    ]

    logger = MemoryLogger()

    for name in attribute_names:
        method = getattr(logger, name)

        assert callable(method)

        for number in range(3):
            method(f'kek_{name}', 'lol', 'cheburek', name, pek='mek', kekokek=name)

    assert len(logger.data.debug) == 3
    assert len(logger.data.info) == 3
    assert len(logger.data.warning) == 3
    assert len(logger.data.error) == 3
    assert len(logger.data.exception) == 3
    assert len(logger.data.critical) == 3

    assert logger.data.debug[0] == logger.data.debug[1] == logger.data.debug[2]
    assert logger.data.info[0] == logger.data.info[1] == logger.data.info[2]
    assert logger.data.warning[0] == logger.data.warning[1] == logger.data.warning[2]
    assert logger.data.error[0] == logger.data.error[1] == logger.data.error[2]
    assert logger.data.exception[0] == logger.data.exception[1] == logger.data.exception[2]
    assert logger.data.critical[0] == logger.data.critical[1] == logger.data.critical[2]

    assert logger.data.debug[0] == LoggerCallData(message='kek_debug', args=('lol', 'cheburek', 'debug'), kwargs={'pek': 'mek', 'kekokek': 'debug'})
    assert logger.data.info[0] == LoggerCallData(message='kek_info', args=('lol', 'cheburek', 'info'), kwargs={'pek': 'mek', 'kekokek': 'info'})
    assert logger.data.warning[0] == LoggerCallData(message='kek_warning', args=('lol', 'cheburek', 'warning'), kwargs={'pek': 'mek', 'kekokek': 'warning'})
    assert logger.data.error[0] == LoggerCallData(message='kek_error', args=('lol', 'cheburek', 'error'), kwargs={'pek': 'mek', 'kekokek': 'error'})
    assert logger.data.exception[0] == LoggerCallData(message='kek_exception', args=('lol', 'cheburek', 'exception'), kwargs={'pek': 'mek', 'kekokek': 'exception'})
    assert logger.data.critical[0] == LoggerCallData(message='kek_critical', args=('lol', 'cheburek', 'critical'), kwargs={'pek': 'mek', 'kekokek': 'critical'})


def test_repr_memory_logger():
    assert repr(MemoryLogger()) == 'MemoryLogger()'



import logging

import pytest
from full_match import match
from loguru import logger as loguru_logger

from emptylog.abstract_logger import AbstractLogger
from emptylog import EmptyLogger, LoggersGroup, MemoryLogger, PrintingLogger


@pytest.mark.parametrize(
    ['first_logger'],
    (
        (EmptyLogger(),),
        (MemoryLogger(),),
        (PrintingLogger(),),
    ),
)
@pytest.mark.parametrize(
    ['second_logger'],
    (
        (EmptyLogger(),),
        (MemoryLogger(),),
        (PrintingLogger(),),
        (logging,),
        (logging.getLogger('kek'),),
        (loguru_logger,),
    ),
)
def test_sum_of_inner_loggers(first_logger, second_logger):
    sum = first_logger + second_logger

    assert isinstance(sum, LoggersGroup)

    assert sum is not first_logger
    assert sum is not second_logger

    assert sum.loggers[0] is first_logger
    assert sum.loggers[1] is second_logger


@pytest.mark.parametrize(
    ['first_logger'],
    (
        (logging,),
        (logging.getLogger('kek'),),
        (loguru_logger,),
    ),
)
@pytest.mark.parametrize(
    ['second_logger'],
    (
        (EmptyLogger(),),
        (MemoryLogger(),),
        (PrintingLogger(),),
    ),
)
def test_sum_with_another_loggers_as_first_operand(first_logger, second_logger):
    sum = first_logger + second_logger

    assert isinstance(sum, LoggersGroup)

    assert sum is not first_logger
    assert sum is not second_logger

    assert sum.loggers[0] is first_logger
    assert sum.loggers[1] is second_logger


@pytest.mark.parametrize(
    ['logger'],
    (
        (EmptyLogger(),),
        (LoggersGroup(),),
        (MemoryLogger(),),
        (PrintingLogger(),),
    ),
)
def test_all_loggers_are_instances_of_abstract_logger(logger):
    assert isinstance(logger, AbstractLogger)


@pytest.mark.parametrize(
    ['logger'],
    (
        (EmptyLogger(),),
        (LoggersGroup(),),
        (MemoryLogger(),),
        (PrintingLogger(),),
    ),
)
@pytest.mark.parametrize(
    ['wrong_operand'],
    (
        (1,),
        ('kek',),
        (None,),
    ),
)
def test_sum_with_wrong_first_operand(logger, wrong_operand):
    with pytest.raises(NotImplementedError, match=match('The addition operation is defined only for loggers.')):
        wrong_operand + logger


@pytest.mark.parametrize(
    ['logger'],
    (
        (EmptyLogger(),),
        (LoggersGroup(),),
        (MemoryLogger(),),
        (PrintingLogger(),),
    ),
)
@pytest.mark.parametrize(
    ['wrong_operand'],
    (
        (1,),
        ('kek',),
        (None,),
    ),
)
def test_sum_with_wrong_second_operand(logger, wrong_operand):
    with pytest.raises(NotImplementedError, match=match('The addition operation is defined only for loggers.')):
        logger + wrong_operand


def test_sum_of_three_loggers():
    first_logger = EmptyLogger()
    second_logger = MemoryLogger()
    third_logger = PrintingLogger()

    sum = first_logger + second_logger + third_logger

    assert isinstance(sum, LoggersGroup)

    assert len(sum) == 3
    assert len(sum.loggers) == 3

    assert sum.loggers[0] is first_logger
    assert sum.loggers[1] is second_logger
    assert sum.loggers[2] is third_logger



import io
import re
from contextlib import redirect_stdout

import pytest

from emptylog import PrintingLogger, LoggerProtocol


def test_printing_logger_is_logger():
    assert isinstance(PrintingLogger(), LoggerProtocol)


@pytest.mark.parametrize(
    ['method', 'result_tail'],
    (
        (PrintingLogger().debug, ' | DEBUG     | kek'),
        (PrintingLogger().info, ' | INFO      | kek'),
        (PrintingLogger().warning, ' | WARNING   | kek'),
        (PrintingLogger().error, ' | ERROR     | kek'),
        (PrintingLogger().exception, ' | EXCEPTION | kek'),
        (PrintingLogger().critical, ' | CRITICAL  | kek'),

        (PrintingLogger(separator='*').debug, ' * DEBUG     * kek'),
        (PrintingLogger(separator='*').info, ' * INFO      * kek'),
        (PrintingLogger(separator='*').warning, ' * WARNING   * kek'),
        (PrintingLogger(separator='*').error, ' * ERROR     * kek'),
        (PrintingLogger(separator='*').exception, ' * EXCEPTION * kek'),
        (PrintingLogger(separator='*').critical, ' * CRITICAL  * kek'),
    ),
)
def test_check_simple_output(method, result_tail):
    # this code is adapted from there: https://stackoverflow.com/a/66683635
    remove_suffix = lambda input_string, suffix: input_string[:-len(suffix)] if suffix and input_string.endswith(suffix) else input_string  # noqa: E731

    buffer = io.StringIO()
    with redirect_stdout(buffer):
        method('kek')

    printed = buffer.getvalue()

    assert printed.endswith(result_tail + '\n')

    time_stamp = remove_suffix(printed, result_tail + '\n') # expected format: 2024-07-08 19:09:48.226667

    assert len(time_stamp.split()) == 2

    date = time_stamp.split()[0]
    time = time_stamp.split()[1]

    assert re.match(r'[\d]{4}-[\d]{2}-[\d]{2}', date) is not None

    time_before_dot = time.split('.')[0]
    time_after_dot = time.split('.')[1]

    assert len(time.split('.')) == 2
    assert re.match(r'[\d]{2}:[\d]{2}:[\d]{2}', time_before_dot) is not None
    assert time_after_dot.isdigit()


@pytest.mark.parametrize(
    ['get_method', 'result_tail'],
    (
        (lambda x: x.debug, ' | DEBUG     | kek'),
        (lambda x: x.info, ' | INFO      | kek'),
        (lambda x: x.warning, ' | WARNING   | kek'),
        (lambda x: x.error, ' | ERROR     | kek'),
        (lambda x: x.exception, ' | EXCEPTION | kek'),
        (lambda x: x.critical, ' | CRITICAL  | kek'),
    ),
)
def test_forward_output(get_method, result_tail):
    # this code is adapted from there: https://stackoverflow.com/a/66683635
    remove_suffix = lambda input_string, suffix: input_string[:-len(suffix)] if suffix and input_string.endswith(suffix) else input_string  # noqa: E731

    counter = 0
    last_line = ''

    def callback(line):
        nonlocal counter
        nonlocal last_line
        counter += 1
        last_line = line

    logger = PrintingLogger(printing_callback=callback)
    method = get_method(logger)

    buffer = io.StringIO()
    with redirect_stdout(buffer):
        method('kek')

    assert buffer.getvalue() == ''

    assert last_line.endswith(result_tail + '\n')

    time_stamp = remove_suffix(last_line, result_tail + '\n') # expected format: 2024-07-08 19:09:48.226667

    assert len(time_stamp.split()) == 2

    date = time_stamp.split()[0]
    time = time_stamp.split()[1]

    assert re.match(r'[\d]{4}-[\d]{2}-[\d]{2}', date) is not None

    time_before_dot = time.split('.')[0]
    time_after_dot = time.split('.')[1]

    assert len(time.split('.')) == 2
    assert re.match(r'[\d]{2}:[\d]{2}:[\d]{2}', time_before_dot) is not None
    assert time_after_dot.isdigit()


@pytest.mark.parametrize(
    ['method', 'result_tail'],
    (
        (PrintingLogger().debug, ' | DEBUG     | kek'),
        (PrintingLogger().info, ' | INFO      | kek'),
        (PrintingLogger().warning, ' | WARNING   | kek'),
        (PrintingLogger().error, ' | ERROR     | kek'),
        (PrintingLogger().exception, ' | EXCEPTION | kek'),
        (PrintingLogger().critical, ' | CRITICAL  | kek'),
    ),
)
def test_multiple_lines(method, result_tail):
    number_of_iterations = 10
    lines = []

    for number in range(number_of_iterations):
        buffer = io.StringIO()
        with redirect_stdout(buffer):
            method('kek')

        lines.append(buffer.getvalue())

    assert len(lines) == number_of_iterations

    assert all([x.endswith(result_tail + '\n') for x in lines])


def test_repr_printing_logger():
    assert repr(PrintingLogger()) == 'PrintingLogger()'
